/*
 * disasm - a MIPS disassembler.
 *
 * Copyright (C) 1996 Kai Harrekilde-Petersen <khp@dolphinics.no>
 *
 * File: insns.c
 *
 * Version: 0.38
 * Date: 22. August 1996
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h> /* Needed for NULL */

#include "disasm.h"

/*
 * Opcodes
 */
insn_st opcodes[64] = {
  {NULL, TYPE_SPECIAL, mips1},	{NULL, TYPE_REGIMM, mips1},
  {"j", TYPE_JUMP, mips1},	{"jal", TYPE_JUMP, mips1},
  {"beq", TYPE_BRA, mips1},	{"bne", TYPE_BRA, mips1},
  {"blez", TYPE_BRA, mips1},	{"bgtz", TYPE_BRA, mips1},
  {"addi", TYPE_IMM, mips1},	{"addiu", TYPE_IMM, mips1},
  {"slti", TYPE_IMMU, mips1},	{"sltiu", TYPE_IMMU, mips1},
  {"andi", TYPE_IMMU, mips1},	{"ori", TYPE_IMMU, mips1},
  {"xori", TYPE_IMMU, mips1},	{"lui", TYPE_IMMU, mips1},
  {"cop0", TYPE_CP, mips1},	{"cop1", TYPE_CP, mips1},
  {"cop2", TYPE_CP, mips1},	{"cop3", TYPE_CP, mips1},
  {"beql", TYPE_BRA, mips2},	{"bnel", TYPE_BRA, mips2},
  {"blezl", TYPE_BRA, mips2},	{"bgtzl", TYPE_BRA, mips2},
  {"daddi", TYPE_IMM, mips3},	{"daddiu", TYPE_IMM, mips3},
  {"ldl", TYPE_LDST, mips2},	{"ldr", TYPE_LDST, mips3},
  {NULL, TYPE_RES, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips1},	{NULL, TYPE_RES, mips0},
  {"lb", TYPE_LDST, mips1},	{"lh", TYPE_LDST, mips1},
  {"lwl", TYPE_LDST, mips1},	{"lw", TYPE_LDST, mips1},
  {"lbu", TYPE_LDST, mips1},	{"lhu", TYPE_LDST, mips1},
  {"lwr", TYPE_LDST, mips1},	{"lwu", TYPE_LDST, mips3},
  {"sb", TYPE_LDST, mips1},	{"sh", TYPE_LDST, mips1},
  {"swl", TYPE_LDST, mips1},	{"sw", TYPE_LDST, mips1},
  {"sdl", TYPE_LDST, mips3},	{"sdr", TYPE_LDST, mips3},
  {"swr", TYPE_LDST, mips1},	{"cache", TYPE_CACHE, mips2}, /* FIXME: is CACHE mips-2 or mips-3? */
  {NULL, TYPE_ISA, mips1},	{"lwc1", TYPE_LDST, mips1}, /* lwc0(mips1)->ll(mips2). Both are LDST */
  {"lwc2", TYPE_LDST, mips1},	{"lwc3", TYPE_LDST, mips1},
  {"lld", TYPE_LDST, mips3},	{"ldc1", TYPE_LDST, mips3},
  {"ldc2", TYPE_LDST, mips3},	{"ld", TYPE_LDST, mips3},
  {NULL, TYPE_ISA, mips1},	{"swc1", TYPE_LDST, mips1}, /* swc0(mips1)->sc(mips2). Both are LDST */
  {"swc2", TYPE_LDST, mips1},	{"swc3", TYPE_LDST, mips1},
  {"scd", TYPE_LDST, mips3},	{"sdc1", TYPE_LDST, mips3},
  {"sdc2", TYPE_LDST, mips3},	{"sd", TYPE_LDST, mips3}
};

insn_st special[64] = {
  {"sll", TYPE_SHIFT, mips1},	{NULL, TYPE_RES, mips0},
  {"srl", TYPE_SHIFT, mips1},	{"sra", TYPE_SHIFT, mips1},
  {"sllv", TYPE_REG, mips1},	{NULL, TYPE_RES, mips0},
  {"srlv", TYPE_REG, mips1},	{"srav", TYPE_REG, mips1},
  {"jr", TYPE_JR, mips1},	{"jalr", TYPE_JR, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"syscall", TYPE_SYSCALL, mips1},	{"break", TYPE_SYSCALL, mips1},
  {NULL, TYPE_RES, mips0},	{"sync", TYPE_NOP, mips2},
  {"mfhi", TYPE_HILO, mips1},	{"mthi", TYPE_HILO, mips1},
  {"mflo", TYPE_HILO, mips1},	{"mtlo", TYPE_HILO, mips1},
  {"dsllv", TYPE_REG, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrlv", TYPE_REG, mips2},	{"dsrav", TYPE_REG, mips3},
  {"mult", TYPE_MOVE, mips1},	{"multu", TYPE_MOVE, mips1},
  {"div", TYPE_MOVE, mips1},	{"divu", TYPE_MOVE, mips1},
  {"dmult", TYPE_MOVE, mips3},	{"dmultu", TYPE_MOVE, mips3},
  {"ddiv", TYPE_MOVE, mips3},	{"ddivu", TYPE_MOVE, mips3},
  {"add", TYPE_REG, mips1},	{"addu", TYPE_REG, mips1},
  {"sub", TYPE_REG, mips1},	{"subu", TYPE_REG, mips1},
  {"and", TYPE_REG, mips1},	{"or", TYPE_REG, mips1},
  {"xor", TYPE_REG, mips1},	{"nor", TYPE_REG, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"slt", TYPE_REG, mips1},	{"sltu", TYPE_REG, mips1},
  {"dadd", TYPE_REG, mips3},	{"daddu", TYPE_REG, mips3},
  {"dsub", TYPE_REG, mips3},	{"dsubu", TYPE_REG, mips3},
  {"tge", TYPE_MOVE, mips2},	{"tgeu", TYPE_MOVE, mips2},
  {"tlt", TYPE_MOVE, mips2},	{"tltu", TYPE_MOVE, mips2},
  {"teq", TYPE_MOVE, mips2},	{NULL, TYPE_RES, mips0},
  {"tne", TYPE_MOVE, mips2},	{NULL, TYPE_RES, mips0},
  {"dsll", TYPE_SHIFT, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrl", TYPE_SHIFT, mips3},	{"dsra", TYPE_SHIFT, mips3},
  {"dsll32", TYPE_SHIFT, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrl32", TYPE_SHIFT, mips3},	{"dsra32", TYPE_SHIFT, mips3}
};

insn_st regimm[32] = {
  {"bltz", TYPE_BRA, mips1},	{"bgez", TYPE_PRINT, mips1},
  {"bltzl", TYPE_BRA, mips2},	{"bgezl", TYPE_PRINT, mips2},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"tgei", TYPE_TRAPI, mips2},	{"tgeiu", TYPE_TRAPI, mips2},
  {"tlti", TYPE_TRAPI, mips2},	{"tltiu", TYPE_TRAPI, mips2},
  {"teqi", TYPE_TRAPI, mips2},	{NULL, TYPE_RES, mips0},
  {"tnei", TYPE_TRAPI, mips2},	{NULL, TYPE_RES, mips0},
  {"bltzal", TYPE_BRA, mips1},	{"bgezal", TYPE_PRINT, mips1},
  {"bltzall", TYPE_BRA, mips2},	{"bgezall", TYPE_PRINT, mips2},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

fixup_t fixup_insns[] = {
  {"lwc0", TYPE_NONE, TYPE_LDST, mips1, 0x1a},
  {"ll", TYPE_NONE, TYPE_LDST, mips2, 0x1a},
  {"ll", TYPE_NONE, TYPE_LDST, mips3, 0x1a},
  {"b", TYPE_REGIMM, TYPE_BRA, mips1, 0x01},
  {"bl", TYPE_REGIMM, TYPE_BRA, mips1, 0x3},
  {"bal", TYPE_REGIMM, TYPE_BRA, mips1, 0x11},
  {"ball", TYPE_REGIMM, TYPE_BRA, mips2, 0x13},
  {"rfe", TYPE_CP, TYPE_NOP, mips1, 0x10},
  {"eret", TYPE_CP, TYPE_NOP, mips2, 0x10},
  {"eret", TYPE_CP, TYPE_NOP, mips3, 0x10},
  {"b", TYPE_NONE, TYPE_BRA, mips1, 0x04},
  {"li", TYPE_NONE, TYPE_IMM, mips1, 0x09},
  {NULL, TYPE_NONE, TYPE_NONE, mips0, 0x00} /* end marker */
};

insn_st COPzrs[16] = {
  {"mf", TYPE_CP, mips1},	{"dmf", TYPE_CP, mips3},
  {"cf", TYPE_CP, mips1},	{NULL, TYPE_INV, mips0},
  {"mt", TYPE_CP, mips1},	{"dmt", TYPE_CP, mips3},
  {"ct", TYPE_CP, mips1},	{NULL, TYPE_INV, mips0},
  {"bc", TYPE_CP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
  /* The second half of this table is CoProc specific */
};
  
insn_st COPzrt[32] = {
  {"bcf", TYPE_CP, mips1},	{"bct", TYPE_CP, mips1},
  {"bcfl", TYPE_CP, mips2},	{"bctl", TYPE_CP, mips2},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0}
};
  
insn_st CP0_opcode[64] = {
  {NULL, TYPE_RES, mips0},	{"tlbr", TYPE_NOP, mips1},
  {"tlbwi", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"tlbwr", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {"tlbp", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"BUG", TYPE_ISA, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"wait", TYPE_NOP, mips3},	{NULL, TYPE_RES, mips0}, /* wait is a R4600 specific insn */
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

insn_st CP1_func[64] = {
  {"add", TYPE_REG, mips1},	{"sub", TYPE_REG, mips1},
  {"mul", TYPE_REG, mips1},	{"div", TYPE_REG, mips1},
  {"sqrt", TYPE_MOVE, mips1},	{"abs", TYPE_MOVE, mips1},
  {"mov", TYPE_MOVE, mips1},	{"neg", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"round.w", TYPE_MOVE,mips1},	{"trunc.w", TYPE_MOVE, mips1},
  {"ceil.w", TYPE_MOVE, mips1},	{"floor.w", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"cvt.s", TYPE_MOVE, mips1},	{"cvt.d", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"cvt.w", TYPE_MOVE, mips1},	{"cvt.l", TYPE_MOVE, mips3}, /* cvt.l is possibly a mips1 insn */
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"c.f", TYPE_BRA, mips1},	{"c.un", TYPE_BRA, mips1},
  {"c.eq", TYPE_BRA, mips1},	{"c.ueq", TYPE_BRA, mips1},
  {"c.olt", TYPE_BRA, mips1},	{"c.ult", TYPE_BRA, mips1},
  {"c.ole", TYPE_BRA, mips1},	{"c.ule", TYPE_BRA, mips1},
  {"c.sf", TYPE_BRA, mips1},	{"c.ngle", TYPE_BRA, mips1},
  {"c.seq", TYPE_BRA, mips1},	{"c.ngl", TYPE_BRA, mips1},
  {"c.lt", TYPE_BRA, mips1},	{"c.nge", TYPE_BRA, mips1},
  {"c.le", TYPE_BRA, mips1},	{"c.ngt", TYPE_BRA, mips1}
};

insn_st CP1_fmt[16] = {
  {"s", TYPE_NOP, mips1},	{"d", TYPE_NOP, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"w", TYPE_NOP, mips1},	{"l", TYPE_NOP, mips3},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

char *r3k_CP0regs[32] = {
  "index",	"random",	"entrylo",	"3",
  "context",	"5",		"6",		"7",
  "bvaddr",	"9",		"entryhi",	"11",
  "sr",		"cause",	"epc",		"prid",
  "16",		"17",		"18",		"19",
  "20",		"21",		"22",		"23",
  "24",		"25",		"26",		"27",
  "28",		"29",		"30",		"31"
};

char *r4k_CP0regs[32] = {
  "index",	"random",	"entrylo",	"3",
  "context",	"pagemask",	"wired",	"error",
  "bvaddr",	"count",	"entryhi",	"compare",
  "sr",		"cause",	"epc",		"prid",
  "config",	"lladdr",	"watchlo",	"watchhi",
  "20",		"21",		"22",		"23",
  "24",		"25",		"ecc",		"cacheerr",
  "taglo",	"taghi",	"errepc",	"31"
};

